import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

from dymos import Trajectory, GaussLobatto, Phase, Radau, run_problem
from dymos.examples.shuttle_reentry.shuttle_ode import ShuttleODE

# Expected results from Betts
expected_results = {'constrained': {'time': 2198.67, 'theta': 30.6255},
                    'unconstrained': {'time': 2008.59, 'theta': 34.1412}}


@use_tempdirs
class TestReentry(unittest.TestCase):

    def make_problem(self, constrained=True, transcription=GaussLobatto, optimizer='SLSQP',
                     numseg=50):
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver(optimizer=optimizer)
        p.driver.declare_coloring()

        if optimizer == 'IPOPT':
            p.driver.opt_settings['max_iter'] = 500
            p.driver.opt_settings['alpha_for_y'] = 'safer-min-dual-infeas'
            p.driver.opt_settings['print_level'] = 0
            p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'
            p.driver.opt_settings['tol'] = 1.0E-7
            p.driver.opt_settings['mu_strategy'] = 'monotone'

        traj = p.model.add_subsystem('traj', Trajectory())
        phase0 = traj.add_phase('phase0',
                                Phase(ode_class=ShuttleODE,
                                      transcription=transcription(num_segments=numseg, order=3)))

        phase0.set_time_options(fix_initial=True, units='s', duration_ref=200)
        phase0.add_state('h', fix_initial=True, fix_final=True, units='ft', rate_source='hdot',
                         targets=['h'], lower=0, ref0=75000, ref=300000, defect_ref=1000)
        phase0.add_state('gamma', fix_initial=True, fix_final=True, units='rad',
                         rate_source='gammadot', targets=['gamma'],
                         lower=-89. * np.pi / 180, upper=89. * np.pi / 180)
        phase0.add_state('phi', fix_initial=True, fix_final=False, units='rad',
                         rate_source='phidot', lower=0, upper=89. * np.pi / 180)
        phase0.add_state('psi', fix_initial=True, fix_final=False, units='rad',
                         rate_source='psidot', targets=['psi'], lower=0, upper=90. * np.pi / 180)
        phase0.add_state('theta', fix_initial=True, fix_final=False, units='rad',
                         rate_source='thetadot', targets=['theta'],
                         lower=-89. * np.pi / 180, upper=89. * np.pi / 180)
        phase0.add_state('v', fix_initial=True, fix_final=True, units='ft/s',
                         rate_source='vdot', targets=['v'], lower=500, ref0=2500, ref=25000)
        phase0.add_control('alpha', units='rad', opt=True,
                           lower=-np.pi / 2, upper=np.pi / 2, targets=['alpha'])
        phase0.add_control('beta', units='rad', opt=True,
                           lower=-89 * np.pi / 180, upper=1 * np.pi / 180, targets=['beta'])

        if constrained:
            phase0.add_path_constraint('q', lower=0, upper=70, ref=70)

        phase0.add_objective('theta', loc='final', ref=-0.01)

        p.setup(check=True, force_alloc_complex=True)

        p.set_val('traj.phase0.states:h',
                  phase0.interp('h', [260000, 80000]), units='ft')
        p.set_val('traj.phase0.states:gamma',
                  phase0.interp('gamma', [-1 * np.pi / 180, -5 * np.pi / 180]),
                  units='rad')
        p.set_val('traj.phase0.states:phi',
                  phase0.interp('phi', [0, 75 * np.pi / 180]), units='rad')
        p.set_val('traj.phase0.states:psi',
                  phase0.interp('psi', [90 * np.pi / 180, 10 * np.pi / 180]),
                  units='rad')
        p.set_val('traj.phase0.states:theta',
                  phase0.interp('theta', [0, 25 * np.pi / 180]), units='rad')
        p.set_val('traj.phase0.states:v', phase0.interp('v', [25600, 2500]), units='ft/s')
        p.set_val('traj.phase0.t_initial', 0, units='s')
        p.set_val('traj.phase0.t_duration', 2000, units='s')
        p.set_val('traj.phase0.controls:alpha',
                  phase0.interp('alpha', [17.4 * np.pi / 180, 17.4 * np.pi / 180]), units='rad')
        p.set_val('traj.phase0.controls:beta',
                  phase0.interp('beta', [-75 * np.pi / 180, 0 * np.pi / 180]), units='rad')

        return p

    @require_pyoptsparse(optimizer='SLSQP')
    def test_partials(self):
        p = self.make_problem(constrained=True, transcription=Radau, optimizer='SLSQP', numseg=5)
        p.run_model()
        cpd = p.check_partials(method='cs', compact_print=True, out_stream=None)
        assert_check_partials(cpd, atol=1.0E-4, rtol=1.1)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_reentry_constrained_radau(self):
        p = self.make_problem(constrained=True, transcription=Radau, optimizer='IPOPT')

        p.run_driver()
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1],
                          expected_results['constrained']['time'],
                          tolerance=1e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.states:theta', units='deg')[-1],
                          expected_results['constrained']['theta'],
                          tolerance=1e-2)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_reentry_constrained_gauss_lobatto(self):
        p = self.make_problem(constrained=True, transcription=GaussLobatto, optimizer='IPOPT')

        p.run_driver()
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1],
                          expected_results['constrained']['time'],
                          tolerance=1e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.states:theta', units='deg')[-1],
                          expected_results['constrained']['theta'],
                          tolerance=1e-2)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_reentry_unconstrained_radau(self):
        p = self.make_problem(constrained=False, transcription=Radau, optimizer='IPOPT')

        p.run_driver()
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1],
                          expected_results['unconstrained']['time'],
                          tolerance=1e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.states:theta', units='deg')[-1],
                          expected_results['unconstrained']['theta'],
                          tolerance=1e-2)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_reentry_unconstrained_gauss_lobatto(self):
        p = self.make_problem(constrained=False, transcription=GaussLobatto, optimizer='IPOPT')

        p.run_driver()
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1],
                          expected_results['unconstrained']['time'],
                          tolerance=1e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.states:theta', units='deg')[-1],
                          expected_results['unconstrained']['theta'],
                          tolerance=1e-2)

    @require_pyoptsparse(optimizer='IPOPT')
    def test_reentry_mixed_controls(self):

        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.declare_coloring(tol=1.0E-12)
        p.driver.options['optimizer'] = 'IPOPT'
        p.driver.opt_settings['alpha_for_y'] = 'safer-min-dual-infeas'
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'
        p.driver.opt_settings['mu_strategy'] = 'monotone'

        traj = p.model.add_subsystem('traj', Trajectory())
        phase0 = traj.add_phase('phase0',
                                Phase(ode_class=ShuttleODE,
                                      transcription=Radau(num_segments=30, order=3)))

        phase0.set_time_options(fix_initial=True, units='s', duration_ref=200)
        phase0.add_state('h', fix_initial=True, fix_final=True, units='ft', rate_source='hdot',
                         targets=['h'], lower=0, ref0=75000, ref=300000, defect_ref=1000)
        phase0.add_state('gamma', fix_initial=True, fix_final=True, units='rad',
                         rate_source='gammadot', targets=['gamma'],
                         lower=-89. * np.pi / 180, upper=89. * np.pi / 180)
        phase0.add_state('phi', fix_initial=True, fix_final=False, units='rad',
                         rate_source='phidot', lower=0, upper=89. * np.pi / 180)
        phase0.add_state('psi', fix_initial=True, fix_final=False, units='rad',
                         rate_source='psidot', targets=['psi'], lower=0,
                         upper=90. * np.pi / 180)
        phase0.add_state('theta', fix_initial=True, fix_final=False, units='rad',
                         rate_source='thetadot', targets=['theta'],
                         lower=-89. * np.pi / 180, upper=89. * np.pi / 180)
        phase0.add_state('v', fix_initial=True, fix_final=True, units='ft/s',
                         rate_source='vdot', targets=['v'], lower=500, ref0=2500, ref=25000)
        phase0.add_control('alpha', units='rad', opt=True,
                           lower=-np.pi / 2, upper=np.pi / 2)
        phase0.add_polynomial_control('beta', order=9, units='rad', opt=True,
                                      lower=-89 * np.pi / 180, upper=1 * np.pi / 180)

        phase0.add_objective('theta', loc='final', ref=-0.01)
        phase0.add_path_constraint('q', lower=0, upper=70, ref=70)

        p.setup(check=True, force_alloc_complex=True)

        p.set_val('traj.phase0.states:h',
                  phase0.interp('h', [260000, 80000]), units='ft')
        p.set_val('traj.phase0.states:gamma',
                  phase0.interp('gamma', [-1 * np.pi / 180, -5 * np.pi / 180]),
                  units='rad')
        p.set_val('traj.phase0.states:phi',
                  phase0.interp('phi', [0, 75 * np.pi / 180]),
                  units='rad')
        p.set_val('traj.phase0.states:psi',
                  phase0.interp('psi', [90 * np.pi / 180, 10 * np.pi / 180]),
                  units='rad')
        p.set_val('traj.phase0.states:theta',
                  phase0.interp('theta', [0, 25 * np.pi / 180]),
                  units='rad')
        p.set_val('traj.phase0.states:v',
                  phase0.interp('v', [25600, 2500]),
                  units='ft/s')
        p.set_val('traj.phase0.t_initial', 0, units='s')
        p.set_val('traj.phase0.t_duration', 2000, units='s')
        p.set_val('traj.phase0.controls:alpha',
                  phase0.interp('alpha', [17.4, 17.4]), units='deg')
        p.set_val('traj.phase0.polynomial_controls:beta',
                  phase0.interp('beta', [-20, 0]), units='deg')

        run_problem(p, simulate=True)

        sol = om.CaseReader('dymos_solution.db').get_case('final')
        sim = om.CaseReader('dymos_simulation.db').get_case('final')

        from scipy.interpolate import interp1d

        t_sol = sol.get_val('traj.phase0.timeseries.time')
        beta_sol = sol.get_val('traj.phase0.timeseries.polynomial_controls:beta', units='deg')

        t_sim = sim.get_val('traj.phase0.timeseries.time')
        beta_sim = sim.get_val('traj.phase0.timeseries.polynomial_controls:beta', units='deg')

        sol_interp = interp1d(t_sol.ravel(), beta_sol.ravel())
        sim_interp = interp1d(t_sim.ravel(), beta_sim.ravel())

        t = np.linspace(0, t_sol.ravel()[-1], 1000)

        assert_near_equal(sim_interp(t), sol_interp(t), tolerance=0.01)

        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1],
                          expected_results['constrained']['time'],
                          tolerance=1e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.states:theta', units='deg')[-1],
                          expected_results['constrained']['theta'],
                          tolerance=1e-2)


if __name__ == '___main__':
    unittest.main()



import numpy as np
from scipy import interpolate

# A set of functions that fit splines to the track.


def get_track_points(track, initial_direction=np.array([1, 0])):
    # given a track description, place nodes along the centerlines in order to fit a spline
    # through them. Nodes are denser around corners
    pos = np.array([0, 0])
    direction = initial_direction

    points = [[0, 0]]

    for i in range(len(track.segments)):
        radius = track.get_segment_radius(i)
        length = track.get_segment_length(i)
        if radius == 0:
            # on a straight
            endpoint = pos + direction * length

            for j in range(1, length.astype(int) - 1):
                if j % 5 == 0:
                    points.append(pos + direction * j)

            pos = endpoint
        else:
            # corner
            # length is sweep in radians
            side = track.get_corner_direction(i)
            if side == 0:
                normal = np.array([-direction[1], direction[0]])
            else:
                normal = np.array([direction[1], -direction[0]])

            xc = pos[0] + radius * normal[0]
            yc = pos[1] + radius * normal[1]
            theta_line = np.arctan2(direction[1], direction[0])
            theta_0 = np.arctan2(pos[1] - yc, pos[0] - xc)
            if side == 0:
                theta_end = theta_0 + length
                direction = np.array(
                    [np.cos(theta_line + length), np.sin(theta_line + length)]
                )
            else:
                theta_end = theta_0 - length
                direction = np.array(
                    [np.cos(theta_line - length), np.sin(theta_line - length)]
                )
            theta_vector = np.linspace(theta_0, theta_end, 100)

            x, y = parametric_circle(theta_vector, xc, yc, radius)

            for j in range(len(x)):
                if j % 10 == 0:
                    points.append([x[j], y[j]])

            pos = np.array([x[-1], y[-1]])

    return np.array(points)


def parametric_circle(t, xc, yc, R):
    x = xc + R * np.cos(t)
    y = yc + R * np.sin(t)
    return x, y


def get_spline(points, interval=0.0001, s=0.0):
    # this function fits the spline
    tck, u = interpolate.splprep(points.transpose(), s=s, k=5)
    unew = np.arange(0, 1.0, interval)
    finespline = interpolate.splev(unew, tck)

    gates = interpolate.splev(u, tck)
    gatesd = interpolate.splev(u, tck, der=1)

    single = interpolate.splev(unew, tck, der=1)
    double = interpolate.splev(unew, tck, der=2)
    curv = (single[0] * double[1] - single[1] * double[0]) / (
        single[0] ** 2 + single[1] ** 2
    ) ** (3 / 2)

    return finespline, gates, gatesd, curv, single


def get_gate_normals(gates, gatesd):
    normals = []
    for i in range(len(gates[0])):
        der = [gatesd[0][i], gatesd[1][i]]
        mag = np.sqrt(der[0] ** 2 + der[1] ** 2)
        normal1 = [-der[1] / mag, der[0] / mag]
        normal2 = [der[1] / mag, -der[0] / mag]

        normals.append([normal1, normal2])

    return normals


def transform_gates(gates):
    # transforms from [[x positions],[y positions]] to [[x0, y0],[x1, y1], etc..]
    newgates = []
    for i in range(len(gates[0])):
        newgates.append(([gates[0][i], gates[1][i]]))
    return newgates


def reverse_transform_gates(gates):
    # transforms from [[x0, y0],[x1, y1], etc..] to [[x positions],[y positions]]
    newgates = np.zeros((2, len(gates)))
    for i in range(len(gates)):
        newgates[0, i] = gates[i][0]
        newgates[1, i] = gates[i][1]
    return newgates


def set_gate_displacements(gate_displacements, gates, normals):
    # does not modify original gates, returns updated version
    newgates = np.copy(gates)
    for i in range(len(gates[0])):
        if i > len(gate_displacements) - 1:
            disp = 0
        else:
            disp = gate_displacements[i]
        # if disp>0:
        normal = normals[i][0]  # always points outwards
        # else:
        # 	normal = normals[i][1] #always points inwards
        newgates[0][i] = newgates[0][i] + disp * normal[0]
        newgates[1][i] = newgates[1][i] + disp * normal[1]
    return newgates



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
import dymos as dm

from dymos.transcriptions.explicit_shooting.test.test_rk_integration_comp import SimpleODE
from dymos.transcriptions.explicit_shooting.ode_evaluation_group import ODEEvaluationGroup


class TestODEEvaluationGroup(unittest.TestCase):

    def test_eval(self):
        ode_class = SimpleODE
        time_options = dm.phase.options.TimeOptionsDictionary()

        time_options['targets'] = 't'
        time_options['units'] = 's'

        state_options = {'x': dm.phase.options.StateOptionsDictionary()}

        state_options['x']['shape'] = (1,)
        state_options['x']['units'] = 's**2'
        state_options['x']['rate_source'] = 'x_dot'
        state_options['x']['targets'] = ['x']

        param_options = {'p': dm.phase.options.ParameterOptionsDictionary()}

        param_options['p']['shape'] = (1,)
        param_options['p']['units'] = 's**2'
        param_options['p']['targets'] = ['p']

        control_options = {}
        polynomial_control_options = {}

        p = om.Problem()

        tx = dm.GaussLobatto(num_segments=1, order=3)

        p.model.add_subsystem('ode_eval', ODEEvaluationGroup(ode_class, time_options, state_options,
                                                             param_options, control_options,
                                                             polynomial_control_options,
                                                             grid_data=tx.grid_data,
                                                             ode_init_kwargs=None))
        p.setup(force_alloc_complex=True)

        p.model.ode_eval.set_segment_index(0)
        p.set_val('ode_eval.states:x', [1.25])
        p.set_val('ode_eval.time', [2.2])
        p.set_val('ode_eval.parameters:p', [1.0])

        p.run_model()

        x = p.get_val('ode_eval.states:x')
        t = p.get_val('ode_eval.time')
        xdot_check = x - t**2 + 1

        assert_near_equal(p.get_val('ode_eval.state_rate_collector.state_rates:x_rate'), xdot_check)

        cpd = p.check_partials(compact_print=True, method='cs')
        assert_check_partials(cpd)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import itertools
import os
import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

import dymos as dm
from dymos.examples.double_integrator.double_integrator_ode import DoubleIntegratorODE
from dymos.utils.testing_utils import assert_timeseries_near_equal


@require_pyoptsparse(optimizer='SLSQP')
def double_integrator_direct_collocation(transcription=dm.GaussLobatto, compressed=True):

    p = om.Problem(model=om.Group())
    p.driver = om.pyOptSparseDriver()
    p.driver.declare_coloring()

    t = transcription(num_segments=30, order=3)

    traj = p.model.add_subsystem('traj', dm.Trajectory())

    phase = traj.add_phase('phase0', dm.Phase(ode_class=DoubleIntegratorODE, transcription=t))

    phase.set_time_options(fix_initial=True, fix_duration=True, units='s')

    phase.add_state('v', fix_initial=True, fix_final=True, rate_source='u', units='m/s')
    phase.add_state('x', fix_initial=True, rate_source='v', units='ft', shape=(1, ))

    phase.add_control('u', units='m/s**2', scaler=0.01, continuity=False, rate_continuity=False,
                      rate2_continuity=False, shape=(1, ), lower=-1.0, upper=1.0)

    phase.set_simulate_options(rtol=1.0E-9, atol=1.0E-9)

    # Maximize distance travelled in one second.
    phase.add_objective('x', loc='final', scaler=-1)

    p.model.linear_solver = om.DirectSolver()

    p.setup(check=True)

    p.set_val('traj.phase0.t_initial', 0.0)
    p.set_val('traj.phase0.t_duration', 1.0)

    p.set_val('traj.phase0.states:x', phase.interp('x', [0, 0.25]), units='m')
    p.set_val('traj.phase0.states:v', phase.interp('v', [0, 0]), units='m/s')
    p.set_val('traj.phase0.controls:u', phase.interp('u', [1, -1]), units='m/s**2')

    dm.run_problem(p, simulate=True, make_plots=True, simulate_kwargs={'times_per_seg': 10})

    return p


@use_tempdirs
class TestDoubleIntegratorExample(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out']:
            if os.path.exists(filename):
                os.remove(filename)

    def _assert_results(self, p, traj=True, tol=1.0E-4):
        if traj:
            x = p.get_val('traj.phase0.timeseries.states:x')
            v = p.get_val('traj.phase0.timeseries.states:v')
        else:
            x = p.get_val('phase0.timeseries.states:x')
            v = p.get_val('phase0.timeseries.states:v')

        assert_near_equal(x[0], 0.0, tolerance=tol)
        assert_near_equal(x[-1], 0.25, tolerance=tol)

        assert_near_equal(v[0], 0.0, tolerance=tol)
        assert_near_equal(v[-1], 0.0, tolerance=tol)

    def test_timeseries_units_gl(self):
        double_integrator_direct_collocation(dm.GaussLobatto, compressed=True)

        sol_case = om.CaseReader('dymos_solution.db').get_case('final')
        sim_case = om.CaseReader('dymos_simulation.db').get_case('final')

        t_sol = sol_case.get_val('traj.phase0.timeseries.time')
        t_sim = sim_case.get_val('traj.phase0.timeseries.time')

        for var in ['states:x', 'states:v', 'state_rates:x', 'state_rates:v', 'controls:u']:
            sol = sol_case.get_val(f'traj.phase0.timeseries.{var}')
            sim = sim_case.get_val(f'traj.phase0.timeseries.{var}')
            assert_timeseries_near_equal(t_sim, sim, t_sol, sol, rel_tolerance=1.0E-3,
                                         abs_tolerance=0.01)

    def test_timeseries_units_radau(self):
        double_integrator_direct_collocation(dm.Radau, compressed=True)

        sol_case = om.CaseReader('dymos_solution.db').get_case('final')
        sim_case = om.CaseReader('dymos_simulation.db').get_case('final')

        t_sol = sol_case.get_val('traj.phase0.timeseries.time')
        t_sim = sim_case.get_val('traj.phase0.timeseries.time')

        for var in ['states:x', 'states:v', 'state_rates:x']:
            sol = sol_case.get_val(f'traj.phase0.timeseries.{var}')
            sim = sim_case.get_val(f'traj.phase0.timeseries.{var}')
            assert_timeseries_near_equal(t_sol, sol, t_sim, sim, rel_tolerance=1.0E-3, abs_tolerance=0.01)



from __future__ import print_function, division, absolute_import

import os
import unittest
import numpy as np

import matplotlib
import matplotlib.pyplot as plt
matplotlib.use('Agg')
plt.style.use('ggplot')

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


tf = np.float128(10)


def solution():
    sqrt_two = np.sqrt(2)
    val = sqrt_two * tf
    c1 = (1.5 * np.exp(-val) - 1) / (np.exp(-val) - np.exp(val))
    c2 = (1 - 1.5 * np.exp(val)) / (np.exp(-val) - np.exp(val))

    ui = c1 * (1 + sqrt_two) + c2 * (1 - sqrt_two)
    uf = c1 * (1 + sqrt_two) * np.exp(val) + c2 * (1 - sqrt_two) * np.exp(-val)
    J = 0.5 * (c1 ** 2 * (1 + sqrt_two) * np.exp(2 * val) + c2 ** 2 * (1 - sqrt_two) * np.exp(-2 * val) -
               (1 + sqrt_two) * c1 ** 2 - (1 - sqrt_two) * c2 ** 2)
    return ui, uf, J


@use_tempdirs
class TestHyperSensitive(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for filename in ['total_coloring.pkl', 'SLSQP.out', 'SNOPT_print.out']:
            if os.path.exists(filename):
                os.remove(filename)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_hyper_sensitive_for_docs(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.plotting import plot_results
        from dymos.examples.hyper_sensitive.hyper_sensitive_ode import HyperSensitiveODE

        # Initialize the problem and assign the driver
        p = om.Problem(model=om.Group())
        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        # Setup the trajectory and its phase
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        transcription = dm.Radau(num_segments=30, order=3, compressed=False)

        phase = traj.add_phase('phase0',
                               dm.Phase(ode_class=HyperSensitiveODE, transcription=transcription))

        phase.set_time_options(fix_initial=True, fix_duration=True)
        phase.add_state('x', fix_initial=True, fix_final=False, rate_source='x_dot', targets=['x'])
        phase.add_state('xL', fix_initial=True, fix_final=False, rate_source='L', targets=['xL'])
        phase.add_control('u', opt=True, targets=['u'])

        phase.add_boundary_constraint('x', loc='final', equals=1)

        phase.add_objective('xL', loc='final')

        p.setup(check=True)

        p.set_val('traj.phase0.states:x', phase.interp('x', [1.5, 1]))
        p.set_val('traj.phase0.states:xL', phase.interp('xL', [0, 1]))
        p.set_val('traj.phase0.t_initial', 0)
        p.set_val('traj.phase0.t_duration', tf)
        p.set_val('traj.phase0.controls:u', phase.interp('u', [-0.6, 2.4]))

        #
        # Solve the problem.
        #
        dm.run_problem(p)

        #
        # Verify that the results are correct.
        #
        ui, uf, J = solution()

        assert_near_equal(p.get_val('traj.phase0.timeseries.controls:u')[0],
                          ui,
                          tolerance=1.5e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.controls:u')[-1],
                          uf,
                          tolerance=1.5e-2)

        assert_near_equal(p.get_val('traj.phase0.timeseries.states:xL')[-1],
                          J,
                          tolerance=1e-2)

        #
        # Simulate the explicit solution and plot the results.
        #
        exp_out = traj.simulate()

        plot_results([('traj.phase0.timeseries.time', 'traj.phase0.timeseries.states:x',
                       'time (s)', 'x $(m)$'),
                      ('traj.phase0.timeseries.time', 'traj.phase0.timeseries.controls:u',
                       'time (s)', 'u $(m/s^2)$')],
                     title='Hyper Sensitive Problem Solution\nRadau Pseudospectral Method',
                     p_sol=p, p_sim=exp_out)

        plt.show()


if __name__ == "__main__":
    unittest.main()



import importlib
import os
import shutil
import unittest
import pathlib

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import require_pyoptsparse
from openmdao.utils.general_utils import set_pyoptsparse_opt
from openmdao.utils.testing_utils import use_tempdirs


import dymos as dm
from dymos.examples.finite_burn_orbit_raise.finite_burn_eom import FiniteBurnODE
from dymos.utils.testing_utils import _get_reports_dir

_, optimizer = set_pyoptsparse_opt('IPOPT', fallback=True)
bokeh_available = importlib.util.find_spec('bokeh') is not None


def make_traj(transcription='gauss-lobatto', transcription_order=3, compressed=False):

    t = {'gauss-lobatto': dm.GaussLobatto(num_segments=5, order=transcription_order, compressed=compressed),
         'radau': dm.Radau(num_segments=20, order=transcription_order, compressed=compressed)}

    traj = dm.Trajectory()

    traj.add_parameter('c', opt=False, val=1.5, units='DU/TU',
                       targets={'burn1': ['c'], 'burn2': ['c']})

    # First Phase (burn)

    burn1 = dm.Phase(ode_class=FiniteBurnODE, transcription=t[transcription])

    burn1 = traj.add_phase('burn1', burn1)

    burn1.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='TU')
    burn1.add_state('r', fix_initial=True, fix_final=False, defect_scaler=100.0,
                    rate_source='r_dot', units='DU')
    burn1.add_state('theta', fix_initial=True, fix_final=False, defect_scaler=100.0,
                    rate_source='theta_dot', units='rad')
    burn1.add_state('vr', fix_initial=True, fix_final=False, defect_scaler=100.0,
                    rate_source='vr_dot', units='DU/TU')
    burn1.add_state('vt', fix_initial=True, fix_final=False, defect_scaler=100.0,
                    rate_source='vt_dot', units='DU/TU')
    burn1.add_state('accel', fix_initial=True, fix_final=False,
                    rate_source='at_dot', units='DU/TU**2')
    burn1.add_state('deltav', fix_initial=True, fix_final=False,
                    rate_source='deltav_dot', units='DU/TU')
    burn1.add_control('u1', rate_continuity=True, rate2_continuity=True, units='deg', scaler=0.01,
                      rate_continuity_scaler=0.001, rate2_continuity_scaler=0.001,
                      lower=-30, upper=30)
    # Second Phase (Coast)
    coast = dm.Phase(ode_class=FiniteBurnODE, transcription=t[transcription])

    coast.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 50), duration_ref=50, units='TU')
    coast.add_state('r', fix_initial=False, fix_final=False, defect_scaler=100.0,
                    rate_source='r_dot', units='DU')
    coast.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                    rate_source='theta_dot', units='rad')
    coast.add_state('vr', fix_initial=False, fix_final=False, defect_scaler=100.0,
                    rate_source='vr_dot', units='DU/TU')
    coast.add_state('vt', fix_initial=False, fix_final=False, defect_scaler=100.0,
                    rate_source='vt_dot', units='DU/TU')
    coast.add_state('accel', fix_initial=True, fix_final=True,
                    rate_source='at_dot', units='DU/TU**2')
    coast.add_state('deltav', fix_initial=False, fix_final=False,
                    rate_source='deltav_dot', units='DU/TU')

    coast.add_parameter('u1', opt=False, val=0.0, units='deg')
    coast.add_parameter('c', val=0.0, units='DU/TU')

    # Third Phase (burn)
    burn2 = dm.Phase(ode_class=FiniteBurnODE, transcription=t[transcription])

    traj.add_phase('coast', coast)
    traj.add_phase('burn2', burn2)

    burn2.set_time_options(initial_bounds=(0.5, 50), duration_bounds=(.5, 10), initial_ref=10, units='TU')
    burn2.add_state('r', fix_initial=False, fix_final=True, defect_scaler=100.0,
                    rate_source='r_dot', targets=['r'], units='DU')
    burn2.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                    rate_source='theta_dot', targets=['theta'], units='rad')
    burn2.add_state('vr', fix_initial=False, fix_final=True, defect_scaler=1000.0,
                    rate_source='vr_dot', targets=['vr'], units='DU/TU')
    burn2.add_state('vt', fix_initial=False, fix_final=True, defect_scaler=1000.0,
                    rate_source='vt_dot', targets=['vt'], units='DU/TU')
    burn2.add_state('accel', fix_initial=False, fix_final=False, defect_scaler=1.0,
                    rate_source='at_dot', targets=['accel'], units='DU/TU**2')
    burn2.add_state('deltav', fix_initial=False, fix_final=False, defect_scaler=1.0,
                    rate_source='deltav_dot', units='DU/TU')

    burn2.add_objective('deltav', loc='final', scaler=100.0)

    burn2.add_control('u1', rate_continuity=True, rate2_continuity=True, units='deg',
                      scaler=0.01, lower=-180, upper=180, targets=['u1'])

    burn1.add_timeseries_output('pos_x')
    coast.add_timeseries_output('pos_x')
    burn2.add_timeseries_output('pos_x')

    burn1.add_timeseries_output('pos_y')
    coast.add_timeseries_output('pos_y')
    burn2.add_timeseries_output('pos_y')

    traj.link_phases(phases=['burn1', 'coast', 'burn2'],
                     vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav'])

    traj.link_phases(phases=['burn1', 'burn2'], vars=['accel'])

    return traj


@require_pyoptsparse(optimizer='SLSQP')
def two_burn_orbit_raise_problem(transcription='gauss-lobatto', optimizer='SLSQP', r_target=3.0,
                                 transcription_order=3, compressed=False,
                                 show_output=True):

    p = om.Problem(model=om.Group())

    p.driver = om.pyOptSparseDriver()
    p.driver.options['optimizer'] = optimizer
    p.driver.declare_coloring()
    if optimizer == 'SNOPT':
        p.driver.opt_settings['Major iterations limit'] = 100
        p.driver.opt_settings['Major feasibility tolerance'] = 1.0E-6
        p.driver.opt_settings['Major optimality tolerance'] = 1.0E-6
        if show_output:
            p.driver.opt_settings['iSumm'] = 6
    elif optimizer == 'IPOPT':
        p.driver.opt_settings['hessian_approximation'] = 'limited-memory'
        p.driver.opt_settings['nlp_scaling_method'] = 'gradient-based'
        p.driver.opt_settings['print_level'] = 0
        p.driver.opt_settings['linear_solver'] = 'mumps'
        p.driver.opt_settings['mu_strategy'] = 'adaptive'

    traj = make_traj(transcription=transcription, transcription_order=transcription_order,
                     compressed=compressed)
    p.model.add_subsystem('traj', subsys=traj)

    # Needed to move the direct solver down into the phases for use with MPI.
    #  - After moving down, used fewer iterations (about 30 less)

    p.setup(check=True)

    # Set Initial Guesses
    p.set_val('traj.parameters:c', value=1.5, units='DU/TU')

    burn1 = p.model.traj.phases.burn1
    burn2 = p.model.traj.phases.burn2
    coast = p.model.traj.phases.coast

    if burn1 in p.model.traj.phases._subsystems_myproc:
        p.set_val('traj.burn1.t_initial', value=0.0)
        p.set_val('traj.burn1.t_duration', value=2.25)

        p.set_val('traj.burn1.states:r', value=burn1.interp('r', [1, 1.5]))
        p.set_val('traj.burn1.states:theta', value=burn1.interp('theta', [0, 1.7]))
        p.set_val('traj.burn1.states:vr', value=burn1.interp('vr', [0, 0]))
        p.set_val('traj.burn1.states:vt', value=burn1.interp('vt', [1, 1]))
        p.set_val('traj.burn1.states:accel', value=burn1.interp('accel', [0.1, 0]))
        p.set_val('traj.burn1.states:deltav', value=burn1.interp('deltav', [0, 0.1]))
        p.set_val('traj.burn1.controls:u1', value=burn1.interp('u1', [-3.5, 13.0]))

    if coast in p.model.traj.phases._subsystems_myproc:
        p.set_val('traj.coast.t_initial', value=2.25)
        p.set_val('traj.coast.t_duration', value=3.0)

        p.set_val('traj.coast.states:r', value=coast.interp('r', [1.3, 1.5]))
        p.set_val('traj.coast.states:theta', value=coast.interp('theta', [2.1767, 1.7]))
        p.set_val('traj.coast.states:vr', value=coast.interp('vr', [0.3285, 0]))
        p.set_val('traj.coast.states:vt', value=coast.interp('vt', [0.97, 1]))
        p.set_val('traj.coast.states:accel', value=coast.interp('accel', [0, 0]))

    if burn2 in p.model.traj.phases._subsystems_myproc:
        p.set_val('traj.burn2.t_initial', value=5.25)
        p.set_val('traj.burn2.t_duration', value=1.75)

        p.set_val('traj.burn2.states:r', value=burn2.interp('r', [1, 3]))
        p.set_val('traj.burn2.states:theta', value=burn2.interp('theta', [0, 4.0]))
        p.set_val('traj.burn2.states:vr', value=burn2.interp('vr', [0, 0]))
        p.set_val('traj.burn2.states:vt', value=burn2.interp('vt', [1, np.sqrt(1 / 3)]))
        p.set_val('traj.burn2.states:accel', value=burn2.interp('accel', [0.1, 0]))
        p.set_val('traj.burn2.states:deltav', value=burn2.interp('deltav', [0.1, 0.2]))
        p.set_val('traj.burn2.controls:u1', value=burn2.interp('u1', [1, 1]))

    dm.run_problem(p, run_driver=False, simulate=True, make_plots=True)

    return p


@use_tempdirs
class TestExampleTwoBurnOrbitRaise(unittest.TestCase):

    def tearDown(self):
        if os.path.isdir('plots'):
            shutil.rmtree('plots')

    @unittest.skipIf(not bokeh_available, 'bokeh unavailable')
    def test_bokeh_plots(self):
        dm.options['plots'] = 'bokeh'

        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer='SLSQP', show_output=False)

        plot_dir = pathlib.Path(_get_reports_dir(p)).joinpath('plots')
        self.assertSetEqual({'plots.html'}, set(os.listdir(plot_dir)))

    def test_mpl_plots(self):
        dm.options['plots'] = 'matplotlib'

        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer='SLSQP', show_output=False)

        expected_files = {'control_rates_u1_rate.png', 'state_rates_r.png', 'states_deltav.png',
                          'states_r.png', 'state_rates_accel.png', 'state_rates_deltav.png',
                          'states_accel.png', 'controls_u1.png', 'states_vr.png', 'pos_x.png',
                          'states_vt.png', 'pos_y.png', 'parameters_u1.png', 'states_theta.png',
                          'control_rates_u1_rate2.png', 'state_rates_vt.png', 'time_phase.png',
                          'parameters_c.png', 'state_rates_theta.png', 'state_rates_vr.png', 'dt_dstau.png'}

        html_files = {str(pathlib.Path(f).with_suffix('.html')) for f in expected_files}
        plot_dir = pathlib.Path(_get_reports_dir(p)).joinpath('plots')
        self.assertSetEqual(expected_files.union(html_files), set(os.listdir(plot_dir)))


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.testing_utils import use_tempdirs

from dymos.utils.testing_utils import assert_check_partials
from dymos.examples.brachistochrone.brachistochrone_vector_states_ode import \
    BrachistochroneVectorStatesODE


@use_tempdirs
class TestBrachistochroneVectorStatesODE(unittest.TestCase):

    def test_partials(self):
        nn = 5

        p = om.Problem(model=om.Group())

        ivc = p.model.add_subsystem('ivc', om.IndepVarComp(), promotes_outputs=['*'])
        ivc.add_output('v', val=np.ones((nn,)), units='m/s')
        ivc.add_output('g', val=np.zeros((nn,)), units='m/s**2')
        ivc.add_output('theta', val=np.zeros((nn)), units='rad')

        p.model.add_subsystem('eom', BrachistochroneVectorStatesODE(num_nodes=nn),
                              promotes_inputs=['*'], promotes_outputs=['*'])
        p.setup(check=True, force_alloc_complex=True)

        p['v'] = np.random.rand(nn)
        p['g'] = np.random.rand(nn)
        p['theta'] = np.random.rand(nn)

        p.run_model()

        np.set_printoptions(linewidth=1024, edgeitems=1000)
        cpd = p.check_partials(method='cs')
        assert_check_partials(cpd)



import unittest
import warnings

import numpy as np

import openmdao.api as om
import dymos as dm

from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE


class Simple2StateODE(om.ExplicitComponent):
    """
    A simple ODE from https://math.okstate.edu/people/yqwang/teaching/math4513_fall11/Notes/rungekutta.pdf
    """
    def initialize(self):
        self.options.declare('num_nodes', types=(int,))

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('x', shape=(nn,), units='s**2')
        self.add_input('y', shape=(nn,), units='s')
        self.add_input('t', shape=(nn,), units='s')
        self.add_input('p', shape=(nn,), units='s**2')

        self.add_output('x_dot', shape=(nn,), units='s')
        self.add_output('y_dot', shape=(nn,), units=None)

        ar = np.arange(nn, dtype=int)
        self.declare_partials(of='x_dot', wrt='x', rows=ar, cols=ar, val=1.0)
        self.declare_partials(of='x_dot', wrt='t', rows=ar, cols=ar)
        self.declare_partials(of='x_dot', wrt='p', rows=ar, cols=ar, val=1.0)

        self.declare_partials(of='y_dot', wrt='y', rows=ar, cols=ar, val=-2.0)
        self.declare_partials(of='y_dot', wrt='t', rows=ar, cols=ar, val=1.0)

    def compute(self, inputs, outputs):
        x = inputs['x']
        y = inputs['y']
        t = inputs['t']
        p = inputs['p']
        outputs['x_dot'] = x - t**2 + p
        outputs['y_dot'] = -2 * y + t**3 * np.exp(-2*t)

    def compute_partials(self, inputs, partials):
        t = inputs['t']
        partials['x_dot', 't'] = -2*t
        partials['y_dot', 't'] = -np.exp(-2*t) * t**2 * (2 * t - 3)


class Simple1StateODE(om.ExplicitComponent):
    """
    A simple ODE from https://math.okstate.edu/people/yqwang/teaching/math4513_fall11/Notes/rungekutta.pdf
    """
    def initialize(self):
        self.options.declare('num_nodes', types=(int,))

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input('y', shape=(nn,), units=None)
        self.add_input('t', shape=(nn,), units='s')

        self.add_output('y_dot', shape=(nn,), units='1/s')

        ar = np.arange(nn, dtype=int)

        self.declare_partials(of='y_dot', wrt='y', rows=ar, cols=ar, val=-2.0)
        self.declare_partials(of='y_dot', wrt='t', rows=ar, cols=ar, val=1.0)

    def compute(self, inputs, outputs):
        y = inputs['y']
        t = inputs['t']
        outputs['y_dot'] = -2 * y + t**3 * np.exp(-2*t)

    def compute_partials(self, inputs, partials):
        t = inputs['t']
        partials['y_dot', 't'] = -np.exp(-2*t) * t**2 * (2 * t - 3)


@use_tempdirs
class TestExplicitShooting(unittest.TestCase):

    def test_1_state_run_model(self):
        prob = om.Problem()

        tx = dm.transcriptions.ExplicitShooting(num_segments=1, grid='gauss-lobatto', method='rk4',
                                                order=3, num_steps_per_segment=10, compressed=True)

        phase = dm.Phase(ode_class=Simple1StateODE, transcription=tx)

        phase.set_time_options(targets=['t'], units='s')

        # automatically discover states
        phase.set_state_options('y', targets=['y'], rate_source='y_dot')

        prob.model.add_subsystem('phase0', phase)

        prob.setup(force_alloc_complex=False)

        prob.set_val('phase0.t_initial', 0.0)
        prob.set_val('phase0.t_duration', 1.0)
        prob.set_val('phase0.states:y', 1.0)

        prob.run_model()

        with np.printoptions(linewidth=1024):
            cpd = prob.check_partials(compact_print=True)

        assert_check_partials(cpd, rtol=1.0E-5)

    def test_2_states_run_model(self):

        for method in ['rk4', 'euler', '3/8', 'ralston', 'rkf', 'rkck', 'dopri']:
            with self.subTest(f"test brachistochrone explicit shooting with method '{method}'"):

                prob = om.Problem()

                tx = dm.transcriptions.ExplicitShooting(num_segments=2, grid='gauss-lobatto',
                                                        method=method, order=3,
                                                        num_steps_per_segment=50, compressed=True)

                phase = dm.Phase(ode_class=Simple2StateODE, transcription=tx)

                phase.set_time_options(targets=['t'], units='s')

                # automatically discover states
                phase.set_state_options('x', targets=['x'], rate_source='x_dot')
                phase.set_state_options('y', targets=['y'], rate_source='y_dot')

                phase.add_parameter('p', targets=['p'])

                prob.model.add_subsystem('phase0', phase)

                prob.setup(force_alloc_complex=True)

                prob.set_val('phase0.t_initial', 0.0)
                prob.set_val('phase0.t_duration', 1.0)
                prob.set_val('phase0.states:x', 0.5)
                prob.set_val('phase0.states:y', 1.0)
                prob.set_val('phase0.parameters:p', 1)

                prob.run_model()

                t_f = prob.get_val('phase0.integrator.t_final')
                x_f = prob.get_val('phase0.integrator.states_out:x')
                y_f = prob.get_val('phase0.integrator.states_out:y')

                if method == 'euler':
                    tol = 5.0E-2
                else:
                    tol = 1.0E-3

                assert_near_equal(t_f, 1.0)
                assert_near_equal(x_f[-1, ...], 2.64085909, tolerance=tol)
                assert_near_equal(y_f[-1, ...], 0.1691691, tolerance=tol)

                with np.printoptions(linewidth=1024):
                    cpd = prob.check_partials(compact_print=True, method='cs')
                    assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_brachistochrone_explicit_shooting(self):

        for method in ['rk4', 'ralston']:
            for compressed in [True, False]:
                with self.subTest(f"test brachistochrone explicit shooting with method '{method}'"):
                    prob = om.Problem()

                    prob.driver = om.pyOptSparseDriver(optimizer='SLSQP')

                    tx = dm.ExplicitShooting(num_segments=3, grid='gauss-lobatto',
                                             method=method, order=5,
                                             num_steps_per_segment=5,
                                             compressed=compressed)

                    phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

                    phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

                    # automatically discover states
                    phase.set_state_options('x', fix_initial=True)
                    phase.set_state_options('y', fix_initial=True)
                    phase.set_state_options('v', fix_initial=True)

                    phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
                    phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9,
                                      ref=90., rate2_continuity=True)

                    phase.add_boundary_constraint('x', loc='final', equals=10.0)
                    phase.add_boundary_constraint('y', loc='final', equals=5.0)

                    prob.model.add_subsystem('phase0', phase)

                    phase.add_objective('time', loc='final')

                    prob.setup(force_alloc_complex=True)

                    prob.set_val('phase0.t_initial', 0.0)
                    prob.set_val('phase0.t_duration', 2)
                    prob.set_val('phase0.states:x', 0.0)
                    prob.set_val('phase0.states:y', 10.0)
                    prob.set_val('phase0.states:v', 1.0E-6)
                    prob.set_val('phase0.parameters:g', 1.0, units='m/s**2')
                    prob.set_val('phase0.controls:theta', phase.interp('theta', ys=[0.01, 90]), units='deg')

                    prob.run_driver()

                    x = prob.get_val('phase0.timeseries.states:x')
                    y = prob.get_val('phase0.timeseries.states:y')
                    t = prob.get_val('phase0.timeseries.time')
                    theta = prob.get_val('phase0.timeseries.controls:theta')
                    theta_rate = prob.get_val('phase0.timeseries.control_rates:theta_rate')
                    theta_rate2 = prob.get_val('phase0.timeseries.control_rates:theta_rate2')

                    assert_near_equal(x[-1, ...], 10.0, tolerance=1.0E-3)
                    assert_near_equal(y[-1, ...], 5.0, tolerance=1.0E-3)
                    assert_near_equal(t[-1, ...], 1.8016, tolerance=1.0E-2)

                    # Test the continuity constraints
                    tol = 1.0E-6
                    assert_near_equal(theta[1:-2:2, ...], theta[2::2, ...], tolerance=tol)
                    assert_near_equal(theta_rate[1:-2:2, ...], theta_rate[2::2, ...], tolerance=tol)
                    assert_near_equal(theta_rate2[1:-2:2, ...], theta_rate2[2::2, ...], tolerance=tol)

                    with np.printoptions(linewidth=1024):
                        cpd = prob.check_partials(compact_print=True, method='cs', out_stream=None)
                        assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_brachistochrone_explicit_shooting_path_constraint(self):

        for method in ['rk4', 'ralston']:
            with self.subTest(f"test brachistochrone explicit shooting with method '{method}'"):
                prob = om.Problem()

                prob.driver = om.pyOptSparseDriver(optimizer='SLSQP')

                tx = dm.transcriptions.ExplicitShooting(num_segments=10, grid='gauss-lobatto', method=method,
                                                        order=3, num_steps_per_segment=5, compressed=True)

                phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

                phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

                # automatically discover states
                phase.set_state_options('x', fix_initial=True)
                phase.set_state_options('y', fix_initial=True)
                phase.set_state_options('v', fix_initial=True)

                phase.add_parameter('g', val=9, units='m/s**2', opt=True, lower=9, upper=9.80665)
                phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

                phase.add_boundary_constraint('x', loc='final', equals=10.0)
                phase.add_boundary_constraint('y', loc='final', equals=5.0)
                phase.add_path_constraint('ydot', lower=-100, upper=0)

                prob.model.add_subsystem('phase0', phase)

                phase.add_objective('time', loc='final')

                prob.setup(force_alloc_complex=True)

                prob.set_val('phase0.t_initial', 0.0)
                prob.set_val('phase0.t_duration', 2)
                prob.set_val('phase0.states:x', 0.0)
                prob.set_val('phase0.states:y', 10.0)
                prob.set_val('phase0.states:v', 1.0E-6)
                prob.set_val('phase0.controls:theta', phase.interp('theta', ys=[0.01, 90]), units='deg')

                prob.run_driver()

                x = prob.get_val('phase0.timeseries.states:x')
                y = prob.get_val('phase0.timeseries.states:y')
                ydot = prob.get_val('phase0.timeseries.ydot')

                assert_near_equal(x[-1, ...], 10.0, tolerance=1.0E-3)
                self.assertTrue(np.all(ydot < 1.0E-6), msg='Not all elements of path constraint satisfied')
                assert_near_equal(y[-1, ...], 5.0, tolerance=1.0E-3)

                with np.printoptions(linewidth=1024):
                    cpd = prob.check_partials(compact_print=False, method='cs', out_stream=None)
                    assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_brachistochrone_explicit_shooting_path_constraint_polynomial_control(self):
        for method in ['euler', '3/8']:
            with self.subTest(f"test brachistochrone explicit shooting with method '{method}'"):
                prob = om.Problem()

                prob.driver = om.pyOptSparseDriver(optimizer='SLSQP')

                tx = dm.transcriptions.ExplicitShooting(num_segments=1, grid='gauss-lobatto', method=method,
                                                        order=3, num_steps_per_segment=50, compressed=True)

                phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

                phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

                # automatically discover states
                phase.set_state_options('x', fix_initial=True)
                phase.set_state_options('y', fix_initial=True)
                phase.set_state_options('v', fix_initial=True)

                phase.add_parameter('g', val=9, units='m/s**2', opt=True, lower=9, upper=9.80665)
                phase.add_polynomial_control('theta', order=5, val=45.0, units='deg', lower=1.0E-6, upper=179.9)

                phase.add_boundary_constraint('x', loc='final', equals=10.0)
                phase.add_boundary_constraint('y', loc='final', equals=5.0)
                phase.add_path_constraint('ydot', lower=-100, upper=0)

                prob.model.add_subsystem('phase0', phase)

                phase.add_objective('time', loc='final')

                prob.setup(force_alloc_complex=True)

                prob.set_val('phase0.t_initial', 0.0)
                prob.set_val('phase0.t_duration', 2)
                prob.set_val('phase0.states:x', 0.0)
                prob.set_val('phase0.states:y', 10.0)
                prob.set_val('phase0.states:v', 1.0E-6)
                prob.set_val('phase0.polynomial_controls:theta', phase.interp('theta', ys=[0.01, 90]), units='deg')

                prob.run_driver()

                x = prob.get_val('phase0.timeseries.states:x')
                y = prob.get_val('phase0.timeseries.states:y')
                ydot = prob.get_val('phase0.timeseries.ydot')

                assert_near_equal(x[-1, ...], 10.0, tolerance=1.0E-3)
                self.assertTrue(np.all(ydot < 1.0E-6), msg='Not all elements of path constraint satisfied')
                assert_near_equal(y[-1, ...], 5.0, tolerance=1.0E-3)

                with np.printoptions(linewidth=1024):
                    cpd = prob.check_partials(compact_print=False, method='cs', out_stream=None)
                    assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_brachistochrone_explicit_shooting_path_constraint_renamed(self):

        for method in ['rk4', 'ralston']:
            with self.subTest(f"test brachistochrone explicit shooting with method '{method}'"):
                prob = om.Problem()

                prob.driver = om.pyOptSparseDriver(optimizer='SLSQP')

                tx = dm.transcriptions.ExplicitShooting(num_segments=10, grid='gauss-lobatto', method=method,
                                                        order=3, num_steps_per_segment=5, compressed=True)

                phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

                phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

                # automatically discover states
                phase.set_state_options('x', fix_initial=True)
                phase.set_state_options('y', fix_initial=True)
                phase.set_state_options('v', fix_initial=True)

                phase.add_parameter('g', val=9.80665, units='m/s**2', opt=False)
                phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

                phase.add_boundary_constraint('x', loc='final', equals=10.0)
                phase.add_boundary_constraint('y', loc='final', equals=5.0)
                phase.add_path_constraint('ydot', constraint_name='foo', lower=-100, upper=0)

                prob.model.add_subsystem('phase0', phase)

                phase.add_objective('time', loc='final')

                prob.setup(force_alloc_complex=True)

                prob.set_val('phase0.t_initial', 0.0)
                prob.set_val('phase0.t_duration', 2)
                prob.set_val('phase0.states:x', 0.0)
                prob.set_val('phase0.states:y', 10.0)
                prob.set_val('phase0.states:v', 1.0E-6)
                prob.set_val('phase0.controls:theta', phase.interp('theta', ys=[0.01, 90]), units='deg')

                prob.run_driver()

                x = prob.get_val('phase0.timeseries.states:x')
                y = prob.get_val('phase0.timeseries.states:y')
                ydot = prob.get_val('phase0.timeseries.foo')

                assert_near_equal(x[-1, ...], 10.0, tolerance=1.0E-3)
                self.assertTrue(np.all(ydot <= 1.0E-6), msg='Not all elements of path constraint satisfied')
                assert_near_equal(y[-1, ...], 5.0, tolerance=1.0E-3)

                with np.printoptions(linewidth=1024):
                    cpd = prob.check_partials(compact_print=True, method='cs', out_stream=None)
                    assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_brachistochrone_explicit_shooting_path_constraint_invalid_renamed(self):
        for method in ['rk4', 'ralston']:
            with self.subTest(f"test brachistochrone explicit shooting with method '{method}'"):
                prob = om.Problem()

                prob.driver = om.pyOptSparseDriver(optimizer='SLSQP')

                tx = dm.transcriptions.ExplicitShooting(num_segments=10, grid='gauss-lobatto', method=method,
                                                        order=3, num_steps_per_segment=5, compressed=True)

                phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

                phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

                # automatically discover states
                phase.set_state_options('x', fix_initial=True)
                phase.set_state_options('y', fix_initial=True)
                phase.set_state_options('v', fix_initial=True)

                phase.add_parameter('g', val=9.80665, units='m/s**2', opt=False)
                phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

                phase.add_boundary_constraint('x', loc='final', equals=10.0)
                phase.add_boundary_constraint('y', loc='final', equals=5.0)
                phase.add_path_constraint('y', constraint_name='foo', lower=5, upper=100)

                prob.model.add_subsystem('phase0', phase)

                phase.add_objective('time', loc='final')

                with warnings.catch_warnings(record=True) as ctx:
                    warnings.simplefilter('always')
                    prob.setup(check=True)

                self.assertIn("<class 'openmdao.utils.om_warnings.UnusedOptionWarning'>: "
                              "Option 'constraint_name' on path constraint y is only valid for "
                              "ODE outputs. The option is being ignored.", [str(w.message) for w in ctx])

    @require_pyoptsparse(optimizer='SLSQP')
    def test_brachistochrone_explicit_shooting_polynomial_control(self):
        prob = om.Problem()

        prob.driver = om.pyOptSparseDriver(optimizer='SLSQP')

        tx = dm.transcriptions.ExplicitShooting(num_segments=3, grid='radau-ps', method='rk4',
                                                order=3, num_steps_per_segment=10, compressed=True)

        phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

        phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

        # automatically discover states
        phase.set_state_options('x', fix_initial=True)
        phase.set_state_options('y', fix_initial=True)
        phase.set_state_options('v', fix_initial=True)

        phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
        phase.add_polynomial_control('theta', order=9, val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

        phase.add_boundary_constraint('x', loc='final', equals=10.0)
        phase.add_boundary_constraint('y', loc='final', equals=5.0)

        prob.model.add_subsystem('phase0', phase)

        phase.add_objective('time', loc='final')

        prob.setup(force_alloc_complex=True)

        prob.set_val('phase0.t_initial', 0.0)
        prob.set_val('phase0.t_duration', 2)
        prob.set_val('phase0.states:x', 0.0)
        prob.set_val('phase0.states:y', 10.0)
        prob.set_val('phase0.states:v', 1.0E-6)
        prob.set_val('phase0.parameters:g', 1.0, units='m/s**2')
        prob.set_val('phase0.polynomial_controls:theta', phase.interp('theta', ys=[0.01, 90]), units='deg')

        prob.run_driver()

        x = prob.get_val('phase0.timeseries.states:x')
        y = prob.get_val('phase0.timeseries.states:y')
        t = prob.get_val('phase0.timeseries.time')
        tp = prob.get_val('phase0.timeseries.time_phase')

        assert_near_equal(x[-1, ...], 10.0, tolerance=1.0E-5)
        assert_near_equal(y[-1, ...], 5.0, tolerance=1.0E-5)
        assert_near_equal(t[-1, ...], 1.8016, tolerance=5.0E-3)
        assert_near_equal(tp[-1, ...], 1.8016, tolerance=5.0E-3)

        with np.printoptions(linewidth=1024):
            cpd = prob.check_partials(compact_print=False, method='cs', out_stream=None)
            assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_explicit_shooting_timeseries_ode_output(self):

        for method in ['rk4', 'ralston']:
            with self.subTest(f"test brachistochrone explicit shooting with method '{method}'"):
                prob = om.Problem()

                prob.driver = om.pyOptSparseDriver(optimizer='SLSQP')

                tx = dm.transcriptions.ExplicitShooting(num_segments=5, grid='gauss-lobatto', method=method,
                                                        order=3, num_steps_per_segment=10, compressed=True)

                phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

                phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

                # automatically discover states
                phase.set_state_options('x', fix_initial=True)
                phase.set_state_options('y', fix_initial=True)
                phase.set_state_options('v', fix_initial=True)

                phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
                phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

                phase.add_boundary_constraint('x', loc='final', equals=10.0)
                phase.add_boundary_constraint('y', loc='final', equals=5.0)

                phase.add_timeseries_output('*')

                prob.model.add_subsystem('phase0', phase)

                phase.add_objective('time', loc='final')

                prob.setup(force_alloc_complex=True)

                prob.set_val('phase0.t_initial', 0.0)
                prob.set_val('phase0.t_duration', 2)
                prob.set_val('phase0.states:x', 0.0)
                prob.set_val('phase0.states:y', 10.0)
                prob.set_val('phase0.states:v', 1.0E-6)
                prob.set_val('phase0.parameters:g', 1.0, units='m/s**2')
                prob.set_val('phase0.controls:theta', phase.interp('theta', ys=[0.01, 90]), units='deg')

                dm.run_problem(prob, run_driver=True)

                x = prob.get_val('phase0.timeseries.states:x')
                y = prob.get_val('phase0.timeseries.states:y')
                v = prob.get_val('phase0.timeseries.states:v')
                theta = prob.get_val('phase0.timeseries.controls:theta', units='rad')
                check = prob.get_val('phase0.timeseries.check')
                t = prob.get_val('phase0.timeseries.time')

                assert_near_equal(x[-1, ...], 10.0, tolerance=1.0E-3)
                assert_near_equal(y[-1, ...], 5.0, tolerance=1.0E-3)
                assert_near_equal(t[-1, ...], 1.8016, tolerance=1.0E-2)
                assert_near_equal(check, v / np.sin(theta))

                with np.printoptions(linewidth=1024):
                    cpd = prob.check_partials(method='cs', out_stream=None)
                    assert_check_partials(cpd, atol=1.0E-5, rtol=1.0E-5)

    @require_pyoptsparse(optimizer='SLSQP')
    def test_explicit_shooting_unknown_timeseries(self):

        for method in ['euler']:
            with self.subTest(f"test brachistochrone explicit shooting with method '{method}'"):
                prob = om.Problem()

                prob.driver = om.pyOptSparseDriver(optimizer='SLSQP')

                tx = dm.transcriptions.ExplicitShooting(num_segments=5, grid='gauss-lobatto', method=method,
                                                        order=3, num_steps_per_segment=10, compressed=True)

                phase = dm.Phase(ode_class=BrachistochroneODE, transcription=tx)

                phase.set_time_options(units='s', fix_initial=True, duration_bounds=(1.0, 10.0))

                # automatically discover states
                phase.set_state_options('x', fix_initial=True)
                phase.set_state_options('y', fix_initial=True)
                phase.set_state_options('v', fix_initial=True)

                phase.add_parameter('g', val=1.0, units='m/s**2', opt=True, lower=1, upper=9.80665)
                phase.add_control('theta', val=45.0, units='deg', opt=True, lower=1.0E-6, upper=179.9)

                phase.add_boundary_constraint('x', loc='final', equals=10.0)
                phase.add_boundary_constraint('y', loc='final', equals=5.0)

                phase.add_timeseries_output('*')
                phase.add_timeseries_output('foo')

                prob.model.add_subsystem('phase0', phase)

                phase.add_objective('time', loc='final')

                msg = "phase0: The following timeseries outputs were requested but not found in the " \
                      "ODE: foo"

                with warnings.catch_warnings(record=True) as ctx:
                    warnings.simplefilter('always')
                    prob.setup(force_alloc_complex=True)

                self.assertIn(msg, [str(w.message) for w in ctx])

    def test_brachistochrone_static_gravity_explicit_shooting(self):
        import openmdao.api as om
        from openmdao.utils.assert_utils import assert_near_equal
        import dymos as dm
        from dymos.examples.brachistochrone.brachistochrone_ode import BrachistochroneODE

        #
        # Initialize the Problem and the optimization driver
        #
        p = om.Problem(model=om.Group())
        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        #
        # Create a trajectory and add a phase to it
        #
        traj = p.model.add_subsystem('traj', dm.Trajectory())

        phase = traj.add_phase('phase0',
                               dm.Phase(ode_class=BrachistochroneODE,
                                        ode_init_kwargs={'static_gravity': True},
                                        transcription=dm.ExplicitShooting(num_segments=10, num_steps_per_segment=10)))

        #
        # Set the variables
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('x', rate_source='xdot',
                        targets=None,
                        units='m',
                        fix_initial=True)

        phase.add_state('y', rate_source='ydot',
                        targets=None,
                        units='m',
                        fix_initial=True)

        phase.add_state('v', rate_source='vdot',
                        targets=['v'],
                        units='m/s',
                        fix_initial=True)

        phase.add_control('theta', targets=['theta'],
                          continuity=True, rate_continuity=True,
                          units='deg', lower=0.01, upper=179.9)

        phase.add_parameter('g', targets=['g'], static_target=True, opt=False)

        #
        # Constrain the final values of x and y
        #
        phase.add_boundary_constraint('x', loc='final', equals=10)
        phase.add_boundary_constraint('y', loc='final', equals=5)

        #
        # Minimize time at the end of the phase
        #
        phase.add_objective('time', loc='final', scaler=10)

        #
        # Setup the Problem
        #
        p.setup()

        #
        # Set the initial values
        # The initial time is fixed, and we set that fixed value here.
        # The optimizer is allowed to modify t_duration, but an initial guess is provided here.
        #
        p.set_val('traj.phase0.t_initial', 0.0)
        p.set_val('traj.phase0.t_duration', 2.0)

        # Guesses for states are provided at all state_input nodes.
        # We use the phase.interpolate method to linearly interpolate values onto the state input nodes.
        # Since fix_initial=True for all states and fix_final=True for x and y, the initial or final
        # values of the interpolation provided here will not be changed by the optimizer.
        p.set_val('traj.phase0.states:x', phase.interp('x', [0, 10]))
        p.set_val('traj.phase0.states:y', phase.interp('y', [10, 5]))
        p.set_val('traj.phase0.states:v', phase.interp('v', [0, 9.9]))

        # Guesses for controls are provided at all control_input node.
        # Here phase.interpolate is used to linearly interpolate values onto the control input nodes.
        p.set_val('traj.phase0.controls:theta', phase.interp('theta', [5, 100.5]))

        # Set the value for gravitational acceleration.
        p.set_val('traj.phase0.parameters:g', 9.80665)

        #
        # Solve for the optimal trajectory
        #
        dm.run_problem(p, simulate=False)

        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-3)



import numpy as np

import openmdao.api as om

from ..options import options as dymos_options
from ..utils.misc import _unspecified


class PhaseLinkageComp(om.ExplicitComponent):
    """
    Component that provides a constraint between end values in two connected phases.

    Provides a 'linkage' capability between two phases to provide
    continuity in states, time, and other variables between two
    phases.

    Conceptually, each linkage can be thought of as a set of compatibility constraints involving
    one or more variables.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('linkages', default=[])
        self._io_names = {}

    def add_linkage_configure(self, lnk):
        """
        Add a linkage constraint to be managed by this component.

        Each constraint equation consists of a variable in the first phase (var_a) at a given
        location in the first phase (loc_a - either 'initial' or 'final') and a variable in the
        second phase (var_b), with corresponding location (loc_b).

        The resulting linkage equation of constraint is:

        C_linkage = sign_a * vars_a + sign_b * vars_b

        Constraining this linkage to a value of zero, with the default signs
        (sign_a = 1, sign_b = -1) will result in the two variables having the same value at the
        given locations.

        Parameters
        ----------
        lnk : LinkageOptionsDictionary
            The linkage options dictionary defining the given linkage constraint.
        """
        if lnk['connected']:
            return

        self.options['linkages'].append(lnk)

        phase_a = lnk['phase_a']
        phase_b = lnk['phase_b']
        shape = lnk['shape']

        var_a = lnk['constraint_name'] if lnk['constraint_name'] else lnk['var_a'].rpartition('.')[-1]
        var_b = lnk['constraint_name'] if lnk['constraint_name'] else lnk['var_b'].rpartition('.')[-1]

        loc_a = lnk['loc_a']
        loc_b = lnk['loc_b']

        units_a = lnk['units_a'] if lnk['units_a'] is not _unspecified else None
        units_b = lnk['units_b'] if lnk['units_b'] is not _unspecified else None
        units = lnk['units'] if lnk['units'] is not _unspecified else None

        if units_a is None or units is None:
            lnk._conv_a, lnk._offset_a = 1.0, 0.0
        else:
            lnk._conv_a, lnk._offset_a = om.unit_conversion(units_a, units)

        if units_b is None or units is None:
            lnk._conv_b, lnk._offset_b = 1.0, 0.0
        else:
            lnk._conv_b, lnk._offset_b = om.unit_conversion(units_b, units)

        input_a = f'{phase_a}:{var_a}'
        input_b = f'{phase_b}:{var_b}'
        output = f'{phase_a}:{var_a}_{loc_a}|{phase_b}:{var_b}_{loc_b}'
        ishape = (2,) + shape

        lnk._input_a = input_a
        lnk._input_b = input_b
        lnk._idxs_a = (0, ...) if loc_a == 'initial' else (-1, ...)
        lnk._idxs_b = (0, ...) if loc_b == 'initial' else (-1, ...)
        lnk._output = output

        try:
            self.add_input(name=input_a, shape=ishape, val=np.zeros(ishape), units=units_a)
        except ValueError as e:
            pass

        try:
            self.add_input(name=input_b, shape=ishape, val=np.zeros(ishape), units=units_b)
        except ValueError as e:
            pass

        self.add_output(name=output, shape=shape, val=np.zeros(shape), units=units)

        if lnk['equals'] is None and lnk['lower'] is None and lnk['upper'] is None:
            lnk['equals'] = 0.0

        self.add_constraint(name=output, equals=lnk['equals'], lower=lnk['lower'],
                            upper=lnk['upper'], ref=lnk['ref'], ref0=lnk['ref0'],
                            scaler=lnk['scaler'], adder=lnk['adder'], linear=lnk['linear'])

        size = np.prod(lnk['shape'])
        rs = np.arange(size)
        cs_a = rs if loc_a == 'initial' else size + rs
        cs_b = rs if loc_b == 'initial' else size + rs

        self.declare_partials(of=output, wrt=input_a, rows=rs, cols=cs_a,
                              val=lnk['sign_a'] * lnk._conv_a)

        self.declare_partials(of=output, wrt=input_b, rows=rs, cols=cs_b,
                              val=lnk['sign_b'] * lnk._conv_b)

    def compute(self, inputs, outputs):
        """
        Compute the linkage constraint values.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        for lnk in self.options['linkages']:
            input_a = lnk._input_a
            input_b = lnk._input_b
            idxs_a = lnk._idxs_a
            idxs_b = lnk._idxs_b
            output = lnk._output

            a_val = lnk['sign_a'] * (inputs[input_a][idxs_a] + lnk._offset_a) * lnk._conv_a
            b_val = lnk['sign_b'] * (inputs[input_b][idxs_b] + lnk._offset_b) * lnk._conv_b

            outputs[output] = a_val + b_val



import numpy as np
import openmdao.api as om

from ..grid_data import GridData
from ...utils.misc import get_rate_units
from ...options import options as dymos_options


class ContinuityCompBase(om.ExplicitComponent):
    """
    ContinuityComp defines constraints to ensure continuity between adjacent segments.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('grid_data', types=GridData,
                             desc='Container object for grid info')

        self.options.declare('state_options', types=dict,
                             desc='Dictionary of state names/options for the phase')

        self.options.declare('control_options', types=dict,
                             desc='Dictionary of control names/options for the phase')

        self.options.declare('time_units', default=None, allow_none=True, types=str,
                             desc='Units of the integration variable')

    def _configure_state_continuity(self):
        state_options = self.options['state_options']
        num_segend_nodes = self.options['grid_data'].subset_num_nodes['segment_ends']
        num_segments = self.options['grid_data'].num_segments
        compressed = self.options['grid_data'].compressed

        if num_segments <= 1 or compressed:
            return

        for state_name, options in state_options.items():
            shape = options['shape'] if options['shape'] is not None else (1, )
            size = np.prod(shape)
            units = options['units']

            self.name_maps[state_name] = {}

            self.name_maps[state_name]['value_names'] = \
                (f'states:{state_name}', f'defect_states:{state_name}')

            self.add_input(name=f'states:{state_name}',
                           shape=(num_segend_nodes,) + shape,
                           desc=f'Values of state {state_name} at segment endpoint nodes',
                           units=units)

            self.add_output(
                name=f'defect_states:{state_name}',
                shape=(num_segments - 1,) + shape,
                desc=f'Consistency constraint values for state {state_name}',
                units=units)

            rs_size1 = np.repeat(np.arange(num_segments - 1, dtype=int), 2)
            cs_size1 = np.arange(1, num_segend_nodes - 1, dtype=int)

            template = np.zeros((num_segments - 1, num_segend_nodes))
            template[rs_size1, cs_size1] = 1.0
            template = np.kron(template, np.eye(size))
            rs, cs = template.nonzero()

            vals = np.zeros(len(rs), dtype=float)
            vals[0::2] = -1.0
            vals[1::2] = 1.0

            self.declare_partials(f'defect_states:{state_name}', f'states:{state_name}',
                                  val=vals, rows=rs, cols=cs)

    def _configure_control_continuity(self):
        control_options = self.options['control_options']
        num_segend_nodes = self.options['grid_data'].subset_num_nodes['segment_ends']
        num_segments = self.options['grid_data'].num_segments
        time_units = self.options['time_units']

        if num_segments <= 1:
            # Control value and rate continuity is enforced even with compressed transcription
            return

        any_rate_continuity = False

        for control_name, options in control_options.items():
            shape = options['shape']
            size = np.prod(shape)
            units = options['units']
            rate_units = get_rate_units(units, time_units, deriv=1)
            rate2_units = get_rate_units(units, time_units, deriv=2)

            # Define the sparsity pattern for rate and rate2 continuity
            rs_size1 = np.repeat(np.arange(num_segments - 1, dtype=int), 2)
            cs_size1 = np.arange(1, num_segend_nodes - 1, dtype=int)

            template = np.zeros((num_segments - 1, num_segend_nodes))
            template[rs_size1, cs_size1] = 1.0
            template = np.kron(template, np.eye(size))
            rs, cs = template.nonzero()

            vals = np.zeros(len(rs), dtype=float)
            vals[0::2] = -1.0
            vals[1::2] = 1.0
            self.rate_jac_templates[control_name] = vals

            #
            # Setup value continuity
            #
            self.name_maps[control_name] = {}

            if options['continuity']:
                self.name_maps[control_name]['value_names'] = \
                    (f'controls:{control_name}',
                     f'defect_controls:{control_name}')

                self.add_input(
                    name=f'controls:{control_name}',
                    shape=(num_segend_nodes,) + shape,
                    desc=f'Values of control {control_name} at segment endpoint nodes',
                    units=units)

                self.add_output(
                    name=f'defect_controls:{control_name}',
                    val=5*np.ones((num_segments - 1,) + shape),
                    desc=f'Continuity constraint values for control {control_name}',
                    units=units)

                self.declare_partials(
                    f'defect_controls:{control_name}',
                    f'controls:{control_name}',
                    val=vals, rows=rs, cols=cs,
                )

            if options['rate_continuity']:
                any_rate_continuity = True

                self.name_maps[control_name]['rate_names'] = \
                    (f'control_rates:{control_name}_rate',
                     f'defect_control_rates:{control_name}_rate')

                self.add_input(
                    name=f'control_rates:{control_name}_rate',
                    shape=(num_segend_nodes,) + shape,
                    desc=f'Values of control {control_name} derivative at segment endpoint nodes',
                    units=rate_units)

                self.add_output(
                    name=f'defect_control_rates:{control_name}_rate',
                    shape=(num_segments - 1,) + shape,
                    desc=f'Consistency constraint values for control {control_name} derivative',
                    units=rate_units)

                self.declare_partials(
                    f'defect_control_rates:{control_name}_rate',
                    f'control_rates:{control_name}_rate',
                    rows=rs, cols=cs,
                )

                self.declare_partials(
                    f'defect_control_rates:{control_name}_rate',
                    't_duration', dependent=True,
                )

            if options['rate2_continuity']:
                any_rate_continuity = True

                self.name_maps[control_name]['rate2_names'] = \
                    (f'control_rates:{control_name}_rate2',
                     f'defect_control_rates:{control_name}_rate2')

                self.add_input(
                    name=f'control_rates:{control_name}_rate2',
                    shape=(num_segend_nodes,) + shape,
                    desc=f'Values of control {control_name} second derivative '
                         'at discretization nodes',
                    units=rate2_units)

                self.add_output(
                    name=f'defect_control_rates:{control_name}_rate2',
                    shape=(num_segments - 1,) + shape,
                    desc='Consistency constraint values for control '
                         f'{control_name} second derivative',
                    units=rate2_units)

                self.declare_partials(
                    f'defect_control_rates:{control_name}_rate2',
                    f'control_rates:{control_name}_rate2',
                    rows=rs, cols=cs,
                )

                self.declare_partials(
                    f'defect_control_rates:{control_name}_rate2',
                    't_duration', dependent=True
                )

        if any_rate_continuity:
            self.add_input('t_duration', units=time_units, val=1.0, desc='time duration of the phase')

    def configure_io(self):
        """
        Configure the input/outputs.

        I/O creation is delayed until configure so that we can determine the shape and units for
        the states.
        """
        self.rate_jac_templates = {}
        self.name_maps = {}

        self._configure_state_continuity()
        self._configure_control_continuity()

    def _compute_state_continuity(self, inputs, outputs):
        state_options = self.options['state_options']
        num_segments = self.options['grid_data'].num_segments
        compressed = self.options['grid_data'].compressed

        if num_segments <= 1 or compressed:
            return

        for state_name in state_options:
            input_name, output_name = self.name_maps[state_name]['value_names']
            end_vals = inputs[input_name][1:-1:2, ...]
            start_vals = inputs[input_name][2:-1:2, ...]
            outputs[output_name] = start_vals - end_vals

    def _compute_control_continuity(self, inputs, outputs):
        control_options = self.options['control_options']

        if any([options['rate_continuity'] or options['rate2_continuity'] for options in control_options.values()]):
            dt_dptau = inputs['t_duration'] / 2.0

        for name, options in control_options.items():
            if options['continuity']:
                input_name, output_name = self.name_maps[name]['value_names']
                end_vals = inputs[input_name][1:-1:2, ...]
                start_vals = inputs[input_name][2:-1:2, ...]
                outputs[output_name] = start_vals - end_vals

            if options['rate_continuity']:
                input_name, output_name = self.name_maps[name]['rate_names']
                end_vals = inputs[input_name][1:-1:2, ...]
                start_vals = inputs[input_name][2:-1:2, ...]
                outputs[output_name] = (start_vals - end_vals) * dt_dptau

            if options['rate2_continuity']:
                input_name, output_name = self.name_maps[name]['rate2_names']
                end_vals = inputs[input_name][1:-1:2, ...]
                start_vals = inputs[input_name][2:-1:2, ...]
                outputs[output_name] = (start_vals - end_vals) * dt_dptau ** 2

    def compute(self, inputs, outputs):
        """
        Compute continuity outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        """
        self._compute_state_continuity(inputs, outputs)
        self._compute_control_continuity(inputs, outputs)

    def compute_partials(self, inputs, partials):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        """
        control_options = self.options['control_options']

        if not any([options['rate_continuity'] or options['rate2_continuity'] for options in control_options.values()]):
            return

        dt_dptau = 0.5 * inputs['t_duration']

        for control_name, options in control_options.items():
            if options['rate_continuity']:
                input_name, output_name = self.name_maps[control_name]['rate_names']
                val = self.rate_jac_templates[control_name]
                partials[output_name, input_name] = val * dt_dptau

                end_vals = inputs[input_name][1:-1:2, ...]
                start_vals = inputs[input_name][2:-1:2, ...]

                partials[output_name, 't_duration'] = 0.5 * (start_vals - end_vals)

            if options['rate2_continuity']:
                input_name, output_name = self.name_maps[control_name]['rate2_names']
                val = self.rate_jac_templates[control_name]
                partials[output_name, input_name] = val * dt_dptau**2

                end_vals = inputs[input_name][1:-1:2, ...]
                start_vals = inputs[input_name][2:-1:2, ...]

                partials[output_name, 't_duration'] = (start_vals - end_vals) * dt_dptau


class GaussLobattoContinuityComp(ContinuityCompBase):
    """
    ContinuityComp defines constraints to ensure continuity between adjacent segments.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def _configure_state_continuity(self):
        state_options = self.options['state_options']
        num_segments = self.options['grid_data'].num_segments
        compressed = self.options['grid_data'].compressed

        if num_segments <= 1:
            return

        super(GaussLobattoContinuityComp, self)._configure_state_continuity()

        for state_name, options in state_options.items():
            if options['continuity'] and not compressed:

                # linear if states are optimized, because they are dvs.
                # but nonlinear if solve_segments, because its like multiple shooting
                is_linear = not options['solve_segments']
                self.add_constraint(name=f'defect_states:{state_name}',
                                    equals=0.0, scaler=1.0, linear=is_linear)

    def _configure_control_continuity(self):
        control_options = self.options['control_options']
        num_segments = self.options['grid_data'].num_segments
        compressed = self.options['grid_data'].compressed

        if num_segments <= 1:
            # Control value and rate continuity is enforced even with compressed transcription
            return

        super(GaussLobattoContinuityComp, self)._configure_control_continuity()

        for control_name, options in control_options.items():

            if options['continuity'] and not compressed:
                self.add_constraint(name=f'defect_controls:{control_name}',
                                    equals=0.0, scaler=1.0, linear=True)

            #
            # Setup first derivative continuity
            #

            if options['rate_continuity']:
                self.add_constraint(name=f'defect_control_rates:{control_name}_rate',
                                    equals=0.0, scaler=options['rate_continuity_scaler'],
                                    linear=False)

            #
            # Setup second derivative continuity
            #

            if options['rate2_continuity']:
                self.add_constraint(name=f'defect_control_rates:{control_name}_rate2',
                                    equals=0.0, scaler=options['rate2_continuity_scaler'],
                                    linear=False)


class RadauPSContinuityComp(ContinuityCompBase):
    """
    ContinuityComp defines constraints to ensure continuity between adjacent segments.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def _configure_state_continuity(self):
        state_options = self.options['state_options']
        num_segments = self.options['grid_data'].num_segments
        compressed = self.options['grid_data'].compressed

        if num_segments <= 1:
            return

        super(RadauPSContinuityComp, self)._configure_state_continuity()

        for state_name, options in state_options.items():
            if options['continuity'] and not compressed:
                # linear if states are optimized, because they are dvs.
                # but nonlinear if solve_segments, because its like multiple shooting
                is_linear = not options['solve_segments']

                self.add_constraint(name=f'defect_states:{state_name}',
                                    equals=0.0, scaler=1.0, linear=is_linear)

    def _configure_control_continuity(self):
        control_options = self.options['control_options']
        num_segments = self.options['grid_data'].num_segments

        if num_segments <= 1:
            # Control value and rate continuity is enforced even with compressed transcription
            return

        super(RadauPSContinuityComp, self)._configure_control_continuity()

        for control_name, options in control_options.items():
            if options['continuity']:
                self.add_constraint(name=f'defect_controls:{control_name}',
                                    equals=0.0, scaler=1.0, linear=False)

            #
            # Setup first derivative continuity
            #

            if options['rate_continuity']:
                self.add_constraint(name=f'defect_control_rates:{control_name}_rate',
                                    equals=0.0, scaler=options['rate_continuity_scaler'],
                                    linear=False)

            #
            # Setup second derivative continuity
            #

            if options['rate2_continuity']:
                self.add_constraint(name=f'defect_control_rates:{control_name}_rate2',
                                    equals=0.0, scaler=options['rate2_continuity_scaler'],
                                    linear=False)
